# -*- coding: utf-8 -*-
"""
Created on  十一月 07 15:52 2017

@author: aeloyq

### Notice:
The image data argumentation methods were copied from Keras
If this reference is illegal, please announce me by email aeloyq@outlook.com
I will deal with it in short time.

"""
import copy
import numpy as np
from nnbuilder.kernel import *

import re
import os
import scipy.ndimage as ndi

try:
    from PIL import Image as pil_image
except ImportError:
    pil_image = None

if pil_image is not None:
    _PIL_INTERPOLATION_METHODS = {
        'nearest': pil_image.NEAREST,
        'bilinear': pil_image.BILINEAR,
        'bicubic': pil_image.BICUBIC,
    }
    # These methods were only introduced in version 3.4.0 (2016).
    if hasattr(pil_image, 'HAMMING'):
        _PIL_INTERPOLATION_METHODS['hamming'] = pil_image.HAMMING
    if hasattr(pil_image, 'BOX'):
        _PIL_INTERPOLATION_METHODS['box'] = pil_image.BOX
    # This method is new in version 1.1.3 (2013).
    if hasattr(pil_image, 'LANCZOS'):
        _PIL_INTERPOLATION_METHODS['lanczos'] = pil_image.LANCZOS


class DataBase(object):
    def __init__(self):
        pass

    def set_model(self, model):
        pass

    def apply(self, data):
        return data

    def apply_(self, data):
        return self.apply(data)

    def get_output(self, output):
        return output

    def get_output_(self, output):
        return self.get_output(output)

    def get_loss(self, loss):
        return loss

    def get_loss_(self, loss):
        return self.get_loss(loss)

    def get_score(self, score):
        return self.get_loss(score)


class CastX(DataBase):
    def __init__(self, type=kernel.config.catX):
        DataBase.__init__(self)
        self.type = type

    def apply(self, data):
        data['X'] = np.asarray(data['X']).astype(self.type).tolist()


class CastY(DataBase):
    def __init__(self, type=kernel.config.catX):
        DataBase.__init__(self)
        self.type = type

    def apply(self, data):
        data['Y'] = np.asarray(data['Y']).astype(self.type).tolist()


class MaskX(DataBase):
    def __init__(self):
        DataBase.__init__(self)
        self.x_mask = None

    def set_model(self, model):
        size = model.X.size
        attr = model.X.attr
        model.model_inputs['X_Mask'] = kernel.placeholder('X_Mask', size, attr)

    def apply(self, data):
        x = data['X']
        maxlen = max([len(d) for d in x])
        x = np.array(x)
        x = np.array(x)
        mask_x = np.ones([len(x), maxlen]).astype(kernel.config.floatX)
        for idx, i in enumerate(x):
            for j in range(len(i), maxlen):
                i.append(np.zeros_like(i[0]).tolist())
                mask_x[idx, j] = 0
        x_new = []
        for idx in range(len(x[0])):
            x_new.append([x[i][idx] for i in range(len(x))])
        mask_x = mask_x.transpose()
        data['X'], data['X_Mask'] = x_new, mask_x


class MaskY(DataBase):
    def __init__(self):
        DataBase.__init__(self)
        self.x_mask = None

    def set_model(self, model):
        size = model.Y.size
        attr = model.Y.attr
        model.model_inputs['Y_Mask'] = kernel.placeholder('Y_Mask', size, attr)

    def apply(self, data):
        y = data['Y']
        maxlen = max([len(d) for d in y])
        y = np.array(y)
        mask_y = np.ones([len(y), maxlen]).astype(kernel.config.floatX)
        for idx, i in enumerate(y):
            for j in range(len(i), maxlen):
                i.append(np.zeros_like(i[0]).tolist())
                mask_y[idx, j] = 0
        y_new = []
        for idx in range(len(y[0])):
            y_new.append([y[i][idx] for i in range(len(y))])
        mask_y = mask_y.transpose()
        data['y'], data['Y_Mask'] = y_new, mask_y

    def get_output(self, output):
        return output

    def get_loss(self, loss):
        return (loss * self.x_mask).sum(0)

    def get_score(self, score):
        return (score * self.x_mask).sum() / self.x_mask.sum()


class EncodeY(DataBase):
    def __init__(self, num):
        DataBase.__init__(self)
        self.num = num

    def apply(self, data):
        y = data['Y']
        enc_y = np.zeros([len(y), self.num], dtype=kernel.config.catX)
        if y.ndim == 1:
            for i, col in enumerate(y):
                enc_y[i, col] = 1
        else:
            for i, row in enumerate(y):
                for col in row:
                    enc_y[i, col] = 1
        data['Y'] = enc_y


class ImArgument2D(DataBase):
    def __init__(self, size, running, running_num, crop=True, flip=True, rotate=True, fill_mode="nearest",
                 cval=0.):
        DataBase.__init__(self)
        self.size = size
        self.crop = crop
        self.flip = flip
        self.rotate = rotate
        self.fill_mode = fill_mode
        self.cval = cval
        self.running = running
        self.running_num = running_num

    def set_model(self, model):
        model.X.size = [None, model.X.size[1]] + list(self.size)
        model.input_layer.out_dim = [model.X.size[1]] + list(self.size)

    def apply(self, data):
        # crop images
        imgs = np.asarray(data['X'])
        if self.rotate:
            nx = []
            for img in imgs:
                nx.append(random_rotation(img, 90, fill_mode=self.fill_mode, cval=self.cval))
            imgs = np.asarray(nx)
        if self.crop:
            nx = []
            for img in imgs:
                nx.append(random_crop(img, self.size))
            imgs = np.asarray(nx)
        if self.flip:
            nx = []
            for img in imgs:
                if np.random.randint(0, 2) == 0:
                    img = flip_axis(img, 1)
                if np.random.randint(0, 2) == 0:
                    img = flip_axis(img, 2)
                nx.append(flip_axis(img, 2))
            imgs = np.asarray(nx)
        data['X'] = imgs

    def apply_(self, data):
        imgs = np.asarray(data['X'])
        nx = []
        for img in imgs:
            nx.append(self.running(img, self.size))
        imgs = np.asarray(nx)
        imgs = imgs.reshape([imgs.shape[0] * imgs.shape[1], imgs.shape[2], imgs.shape[3], imgs.shape[4]])
        data['X'] = imgs

    def get_output_(self, output):
        batch_size = output.size[0]
        if batch_size is not None:
            batch_size = batch_size / self.running_num
        output = output.reshape(
            [self.running_num, output.shape[0] / self.running_num, output.shape[1]],
            [self.running_num, batch_size, output.size[1]],
            ['search', 'batch', 'unit'])
        output = output.mean(0)
        return output


def tencrop(img, size):
    "Four corner patches and center crop from image and its horizontal reflection"
    r, c = len(img[0]), len(img[0][0])
    rr, cc = size[0], size[1]
    nx = []
    for i in (img, flip_axis(img, 2)):
        nx.append(i[:, 0:rr, 0:cc])
        nx.append(i[:, 0:rr, c - cc:c])
        nx.append(i[:, r - rr:r, 0:cc])
        nx.append(i[:, r - rr:r, c - cc:c])
        rrr = (r - rr) // 2
        ccc = (c - cc) // 2
        nx.append(i[:, rrr:rrr + rr, ccc:ccc + cc])
    nx = np.asarray(nx)
    return nx


def twentycrop(img, size):
    "Four corner patches and center crop from image and its two reflections"
    r, c = len(img[1]), len(img[2])
    rr, cc = size[0], size[1]
    nx = []
    for i in (img, flip_axis(img, 1), flip_axis(img, 2), flip_axis(flip_axis(img, 1), 2)):
        nx.append(i[:, 0:rr, 0:cc])
        nx.append(i[:, 0:rr, c - cc:c])
        nx.append(i[:, r - rr:r, 0:cc])
        nx.append(i[:, r - rr:r, c - cc:c])
        rrr = (r - rr) // 2
        ccc = (c - cc) // 2
        nx.append(i[:, rrr:rrr + rr, ccc:ccc + cc])
    nx = np.asarray(nx)
    return nx


def random_rotation(x, rg, row_axis=1, col_axis=2, channel_axis=0,
                    fill_mode='nearest', cval=0.):
    """Performs a random rotation of a Numpy image tensor.
    # Arguments
        x: Input tensor. Must be 3D.
        rg: Rotation range, in degrees.
        row_axis: Index of axis for rows in the input tensor.
        col_axis: Index of axis for columns in the input tensor.
        channel_axis: Index of axis for channels in the input tensor.
        fill_mode: Points outside the boundaries of the input
            are filled according to the given mode
            (one of `{'constant', 'nearest', 'reflect', 'wrap'}`).
        cval: Value used for points outside the boundaries
            of the input if `mode='constant'`.
    # Returns
        Rotated Numpy image tensor.
    """
    theta = np.pi / 180 * np.random.uniform(-rg, rg)
    rotation_matrix = np.array([[np.cos(theta), -np.sin(theta), 0],
                                [np.sin(theta), np.cos(theta), 0],
                                [0, 0, 1]])

    h, w = x.shape[row_axis], x.shape[col_axis]
    transform_matrix = transform_matrix_offset_center(rotation_matrix, h, w)
    x = apply_transform(x, transform_matrix, channel_axis, fill_mode, cval)
    return x


def random_shift(x, wrg, hrg, row_axis=1, col_axis=2, channel_axis=0,
                 fill_mode='nearest', cval=0.):
    """Performs a random spatial shift of a Numpy image tensor.
    # Arguments
        x: Input tensor. Must be 3D.
        wrg: Width shift range, as a float fraction of the width.
        hrg: Height shift range, as a float fraction of the height.
        row_axis: Index of axis for rows in the input tensor.
        col_axis: Index of axis for columns in the input tensor.
        channel_axis: Index of axis for channels in the input tensor.
        fill_mode: Points outside the boundaries of the input
            are filled according to the given mode
            (one of `{'constant', 'nearest', 'reflect', 'wrap'}`).
        cval: Value used for points outside the boundaries
            of the input if `mode='constant'`.
    # Returns
        Shifted Numpy image tensor.
    """
    h, w = x.shape[row_axis], x.shape[col_axis]
    tx = np.random.uniform(-hrg, hrg) * h
    ty = np.random.uniform(-wrg, wrg) * w
    translation_matrix = np.array([[1, 0, tx],
                                   [0, 1, ty],
                                   [0, 0, 1]])

    transform_matrix = translation_matrix  # no need to do offset
    x = apply_transform(x, transform_matrix, channel_axis, fill_mode, cval)
    return x


def random_shear(x, intensity, row_axis=1, col_axis=2, channel_axis=0,
                 fill_mode='nearest', cval=0.):
    """Performs a random spatial shear of a Numpy image tensor.
    # Arguments
        x: Input tensor. Must be 3D.
        intensity: Transformation intensity.
        row_axis: Index of axis for rows in the input tensor.
        col_axis: Index of axis for columns in the input tensor.
        channel_axis: Index of axis for channels in the input tensor.
        fill_mode: Points outside the boundaries of the input
            are filled according to the given mode
            (one of `{'constant', 'nearest', 'reflect', 'wrap'}`).
        cval: Value used for points outside the boundaries
            of the input if `mode='constant'`.
    # Returns
        Sheared Numpy image tensor.
    """
    shear = np.random.uniform(-intensity, intensity)
    shear_matrix = np.array([[1, -np.sin(shear), 0],
                             [0, np.cos(shear), 0],
                             [0, 0, 1]])

    h, w = x.shape[row_axis], x.shape[col_axis]
    transform_matrix = transform_matrix_offset_center(shear_matrix, h, w)
    x = apply_transform(x, transform_matrix, channel_axis, fill_mode, cval)
    return x


def random_crop(x, size, row_axis=1, col_axis=2):
    r = np.random.randint(0, x.shape[row_axis] - size[0] + 1)
    c = np.random.randint(0, x.shape[col_axis] - size[1] + 1)
    x = x[:, r:r + size[0], c:c + size[1]]
    return x


def random_zoom(x, zoom_range, row_axis=1, col_axis=2, channel_axis=0,
                fill_mode='nearest', cval=0.):
    """Performs a random spatial zoom of a Numpy image tensor.
    # Arguments
        x: Input tensor. Must be 3D.
        zoom_range: Tuple of floats; zoom range for width and height.
        row_axis: Index of axis for rows in the input tensor.
        col_axis: Index of axis for columns in the input tensor.
        channel_axis: Index of axis for channels in the input tensor.
        fill_mode: Points outside the boundaries of the input
            are filled according to the given mode
            (one of `{'constant', 'nearest', 'reflect', 'wrap'}`).
        cval: Value used for points outside the boundaries
            of the input if `mode='constant'`.
    # Returns
        Zoomed Numpy image tensor.
    # Raises
        ValueError: if `zoom_range` isn't a tuple.
    """
    if len(zoom_range) != 2:
        raise ValueError('`zoom_range` should be a tuple or list of two floats. '
                         'Received arg: ', zoom_range)

    if zoom_range[0] == 1 and zoom_range[1] == 1:
        zx, zy = 1, 1
    else:
        zx, zy = np.random.uniform(zoom_range[0], zoom_range[1], 2)
    zoom_matrix = np.array([[zx, 0, 0],
                            [0, zy, 0],
                            [0, 0, 1]])

    h, w = x.shape[row_axis], x.shape[col_axis]
    transform_matrix = transform_matrix_offset_center(zoom_matrix, h, w)
    x = apply_transform(x, transform_matrix, channel_axis, fill_mode, cval)
    return x


def apply_transform(x, transform_matrix, channel_axis=0, fill_mode='nearest', cval=0.):
    """Apply the image transformation specified by a matrix.
    # Arguments
        x: 2D numpy array, single image.
        transform_matrix: Numpy array specifying the geometric transformation.
        channel_axis: Index of axis for channels in the input tensor.
        fill_mode: Points outside the boundaries of the input
            are filled according to the given mode
            (one of `{'constant', 'nearest', 'reflect', 'wrap'}`).
        cval: Value used for points outside the boundaries
            of the input if `mode='constant'`.
    # Returns
        The transformed version of the input.
    """
    x = np.rollaxis(x, channel_axis, 0)
    final_affine_matrix = transform_matrix[:2, :2]
    final_offset = transform_matrix[:2, 2]
    channel_images = [ndi.interpolation.affine_transform(
        x_channel,
        final_affine_matrix,
        final_offset,
        order=0,
        mode=fill_mode,
        cval=cval) for x_channel in x]
    x = np.stack(channel_images, axis=0)
    x = np.rollaxis(x, 0, channel_axis + 1)
    return x


def random_channel_shift(x, intensity, channel_axis=0):
    x = np.rollaxis(x, channel_axis, 0)
    min_x, max_x = np.min(x), np.max(x)
    channel_images = [np.clip(x_channel + np.random.uniform(-intensity, intensity), min_x, max_x)
                      for x_channel in x]
    x = np.stack(channel_images, axis=0)
    x = np.rollaxis(x, 0, channel_axis + 1)
    return x


def transform_matrix_offset_center(matrix, x, y):
    o_x = float(x) / 2 + 0.5
    o_y = float(y) / 2 + 0.5
    offset_matrix = np.array([[1, 0, o_x], [0, 1, o_y], [0, 0, 1]])
    reset_matrix = np.array([[1, 0, -o_x], [0, 1, -o_y], [0, 0, 1]])
    transform_matrix = np.dot(np.dot(offset_matrix, matrix), reset_matrix)
    return transform_matrix


def flip_axis(x, axis):
    x = np.asarray(x).swapaxes(axis, 0)
    x = x[::-1, ...]
    x = x.swapaxes(0, axis)
    return x


def array_to_img(x, data_format=None, scale=True):
    """Converts a 3D Numpy array to a PIL Image instance.
    # Arguments
        x: Input Numpy array.
        data_format: Image data format.
        scale: Whether to rescale image values
            to be within [0, 255].
    # Returns
        A PIL Image instance.
    # Raises
        ImportError: if PIL is not available.
        ValueError: if invalid `x` or `data_format` is passed.
    """
    if pil_image is None:
        raise ImportError('Could not import PIL.Image. '
                          'The use of `array_to_img` requires PIL.')
    x = np.asarray(x, dtype=kernel.config.floatX)
    if x.ndim != 3:
        raise ValueError('Expected image array to have rank 3 (single image). '
                         'Got array with shape:', x.shape)

    if data_format is None:
        data_format = 'jpg'
    if data_format not in {'channels_first', 'channels_last'}:
        raise ValueError('Invalid data_format:', data_format)

    # Original Numpy array x has format (height, width, channel)
    # or (channel, height, width)
    # but target PIL image has format (width, height, channel)
    if data_format == 'channels_first':
        x = x.transpose(1, 2, 0)
    if scale:
        x = x + max(-np.min(x), 0)
        x_max = np.max(x)
        if x_max != 0:
            x /= x_max
        x *= 255
    if x.shape[2] == 3:
        # RGB
        return pil_image.fromarray(x.astype('uint8'), 'RGB')
    elif x.shape[2] == 1:
        # grayscale
        return pil_image.fromarray(x[:, :, 0].astype('uint8'), 'L')
    else:
        raise ValueError('Unsupported channel number: ', x.shape[2])


def img_to_array(img, data_format=None):
    """Converts a PIL Image instance to a Numpy array.
    # Arguments
        img: PIL Image instance.
        data_format: Image data format.
    # Returns
        A 3D Numpy array.
    # Raises
        ValueError: if invalid `img` or `data_format` is passed.
    """
    if data_format is None:
        data_format = 'jpg'
    if data_format not in {'channels_first', 'channels_last'}:
        raise ValueError('Unknown data_format: ', data_format)
    # Numpy array x has format (height, width, channel)
    # or (channel, height, width)
    # but original PIL image has format (width, height, channel)
    x = np.asarray(img, dtype=kernel.config.floatX)
    if len(x.shape) == 3:
        if data_format == 'channels_first':
            x = x.transpose(2, 0, 1)
    elif len(x.shape) == 2:
        if data_format == 'channels_first':
            x = x.reshape((1, x.shape[0], x.shape[1]))
        else:
            x = x.reshape((x.shape[0], x.shape[1], 1))
    else:
        raise ValueError('Unsupported image shape: ', x.shape)
    return x


def load_img(path, grayscale=False, target_size=None, interpolation='nearest'):
    """Loads an image into PIL format.
    # Arguments
        path: Path to image file
        grayscale: Boolean, whether to load the image as grayscale.
        target_size: Either `None` (default to original size)
            or tuple of ints `(img_height, img_width)`.
        interpolation: Interpolation method used to resample the image if the
            target size is different from that of the loaded image.
            Supported methods are "nearest", "bilinear", and "bicubic".
            If PIL version 1.1.3 or newer is installed, "lanczos" is also
            supported. If PIL version 3.4.0 or newer is installed, "box" and
            "hamming" are also supported. By default, "nearest" is used.
    # Returns
        A PIL Image instance.
    # Raises
        ImportError: if PIL is not available.
        ValueError: if interpolation method is not supported.
    """
    if pil_image is None:
        raise ImportError('Could not import PIL.Image. '
                          'The use of `array_to_img` requires PIL.')
    img = pil_image.open(path)
    if grayscale:
        if img.mode != 'L':
            img = img.convert('L')
    else:
        if img.mode != 'RGB':
            img = img.convert('RGB')
    if target_size is not None:
        width_height_tuple = (target_size[1], target_size[0])
        if img.size != width_height_tuple:
            if interpolation not in _PIL_INTERPOLATION_METHODS:
                raise ValueError(
                    'Invalid interpolation method {} specified. Supported '
                    'methods are {}'.format(
                        interpolation,
                        ", ".join(_PIL_INTERPOLATION_METHODS.keys())))
            resample = _PIL_INTERPOLATION_METHODS[interpolation]
            img = img.resize(width_height_tuple, resample)
    return img


def list_pictures(directory, ext='jpg|jpeg|bmp|png|ppm'):
    return [os.path.join(root, f)
            for root, _, files in os.walk(directory) for f in files
            if re.match(r'([\w]+\.(?:' + ext + '))', f)]
